
import numpy as np
import cv2


def fit_affine(src, dst):
    """ Fit an affine matrix given matching features.
            src is nx2 containing (y,x) coordinates in image 1
            src is nx2, with coordinates of matching points in image 2
    """
    n = src.shape[0]

    # setup A (uses src coords)
    A = np.zeros((2*n, 6))

    # fill even rows: [ x1 y1 1 0 0 0 ]
    A[0::2, 0:2] = src
    A[0::2,2] = 1

    # fill odd rows: [0 0 0 x1 y1 1]
    A[1::2, 3:5] = src
    A[1::2,5] = 1

    # b is [x1', y1', x2', y2', ...]:
    b = np.ravel(dst)

    # solve the system:
    t, resid, rank, s  = np.linalg.lstsq(A, b, rcond=None) # t is 6d column vector

    # make the output matrix
    T = np.zeros((3,3))
    T[0,:] = t[0:3]
    T[1,:] = t[3:6]
    T[2,2] = 1

    return T

def save_feat_img(img, feats, filename):
    """ save an image with feature points drawn as red boxes """
    for i in range(dst.shape[0]):
        x = int(feats[i,0])
        y = int(feats[i,1])
        img[y-2:y+4, x-2:x+4, :] = [0,0,1]

    cv2.imwrite(filename, (img*255).astype(np.uint8))

if __name__ == "__main__":

    # actual transformation I1 -> I2 coords
    T_ideal = np.array([
        [1.4, -.01, -18],
        [.1, 0.9, 10],
        [0, 0, 1]])


    noise_variance = 0.2

    # load our good friend van
    van = cv2.imread('van.png').astype(np.float32) / 255.0;
    h, w = van.shape[:2]


    # generate some random matches (IRL, a feature matcher does this)
    src = np.random.randint(100, 400, (20,2))


    # put them into homogeneous coordinates
    src_homog = np.pad(src.T, ((0,1), (0,0)), 'constant', constant_values=1.0)

    # apply T_ideal to get destination match coordinates
    dst_homog = np.dot(T_ideal, src_homog)

    # introduce an outlier match:
    dst_homog[:,0] = [100, 100, 1]

    # add some gaussian noise with sigma = noise_variance pixels:
    dst_homog += noise_variance * np.random.randn(*(dst_homog.shape))

    # slice out non-homogeneous coords (no need to divide by w)
    dst = dst_homog[:2,:].T

    # warp image according to T_ideal
    van_warp = cv2.warpAffine(van, T_ideal[:2,:], (w, h))

    T_fit = fit_affine(src, dst)
    print T_fit

    # warp dst back to original coords using T_fit's inverse
    dst_unwarp = np.dot(np.linalg.inv(T_fit), dst_homog)[:2,:].T
    van_warpfit = cv2.warpAffine(van_warp, np.linalg.inv(T_fit)[:2,:], (w, h))

    # show feature points on original, warped, and unwarped images:
    save_feat_img(van.copy(), src, 'van_orig.png')
    save_feat_img(van_warp.copy(), dst, 'van_warp.png')
    save_feat_img(van_warpfit, dst_unwarp, 'van_warpfit.png')
